#ifndef __SHADER_STATS_H
#define __SHADER_STATS_H
#include <stdio.h>
#include <stdint.h>
#include "util/u_debug.h"

struct agx2_stats {
   int32_t instrs;
   int32_t alu;
   int32_t fscib;
   int32_t ic;
   int32_t code_size;
   int16_t gprs;
   int16_t uniforms;
   int32_t scratch;
   int16_t threads;
   int32_t loops;
   int32_t spills;
   int32_t fills;
};

static inline int
agx2_stats_fprintf(FILE *fp, const char *prefix, const struct agx2_stats *stats)
{
   return fprintf(fp, "%s shader: %u instrs, %u alu, %u fscib, %u ic, %u code size, %u gprs, %u uniforms, %u scratch, %u threads, %u loops, %u:%u spills:fills\n", prefix, stats->instrs, stats->alu, stats->fscib, stats->ic, stats->code_size, stats->gprs, stats->uniforms, stats->scratch, stats->threads, stats->loops, stats->spills, stats->fills);
}

static inline void
agx2_stats_util_debug(struct util_debug_callback *debug, const char *prefix, const struct agx2_stats *stats)
{
   util_debug_message(debug, SHADER_INFO, "%s shader: %u instrs, %u alu, %u fscib, %u ic, %u code size, %u gprs, %u uniforms, %u scratch, %u threads, %u loops, %u:%u spills:fills", prefix, stats->instrs, stats->alu, stats->fscib, stats->ic, stats->code_size, stats->gprs, stats->uniforms, stats->scratch, stats->threads, stats->loops, stats->spills, stats->fills);
}

#define vk_add_agx2_stats(out, stats) do { \
   vk_add_exec_statistic_u64(out, "Instructions", "Instruction count", (stats)->instrs); \
   vk_add_exec_statistic_u64(out, "ALU", "Estimated ALU cycle count", (stats)->alu); \
   vk_add_exec_statistic_u64(out, "FSCIB", "Estimated F16/F32/SCIB cycle count", (stats)->fscib); \
   vk_add_exec_statistic_u64(out, "IC", "Estimated IC cycle count", (stats)->ic); \
   vk_add_exec_statistic_u64(out, "Code size", "Binary size in bytes", (stats)->code_size); \
   vk_add_exec_statistic_u64(out, "GPRs", "Number of 16-bit GPRs", (stats)->gprs); \
   vk_add_exec_statistic_u64(out, "Uniforms", "Number of 16-bit uniform registers", (stats)->uniforms); \
   vk_add_exec_statistic_u64(out, "Scratch", "Scratch size per thread in bytes", (stats)->scratch); \
   vk_add_exec_statistic_u64(out, "Threads", "Maximum number of threads in flight on a compute unit", (stats)->threads); \
   vk_add_exec_statistic_u64(out, "Loops", "Number of hardware loops", (stats)->loops); \
   vk_add_exec_statistic_u64(out, "Spills", "Number of spill (stack store) instructions", (stats)->spills); \
   vk_add_exec_statistic_u64(out, "Fills", "Number of fill (stack load) instructions", (stats)->fills); \
} while(0)

struct midgard_stats {
   int32_t inst;
   int32_t bundles;
   int16_t registers;
   int16_t threads;
   int32_t quadwords;
   int32_t loops;
   int32_t spills;
   int32_t fills;
};

static inline int
midgard_stats_fprintf(FILE *fp, const char *prefix, const struct midgard_stats *stats)
{
   return fprintf(fp, "%s shader: %u inst, %u bundles, %u registers, %u threads, %u quadwords, %u loops, %u:%u spills:fills\n", prefix, stats->inst, stats->bundles, stats->registers, stats->threads, stats->quadwords, stats->loops, stats->spills, stats->fills);
}

static inline void
midgard_stats_util_debug(struct util_debug_callback *debug, const char *prefix, const struct midgard_stats *stats)
{
   util_debug_message(debug, SHADER_INFO, "%s shader: %u inst, %u bundles, %u registers, %u threads, %u quadwords, %u loops, %u:%u spills:fills", prefix, stats->inst, stats->bundles, stats->registers, stats->threads, stats->quadwords, stats->loops, stats->spills, stats->fills);
}

#define vk_add_midgard_stats(out, stats) do { \
   vk_add_exec_statistic_u64(out, "Instructions", "Instruction count", (stats)->inst); \
   vk_add_exec_statistic_u64(out, "Bundles", "Instruction bundles", (stats)->bundles); \
   vk_add_exec_statistic_u64(out, "Registers", "Register usage in vec4s", (stats)->registers); \
   vk_add_exec_statistic_u64(out, "Threads", "Maximum number of threads in flight on a compute unit", (stats)->threads); \
   vk_add_exec_statistic_u64(out, "Quadwords", "Binary size in quadwords", (stats)->quadwords); \
   vk_add_exec_statistic_u64(out, "Loops", "Number of hardware loops", (stats)->loops); \
   vk_add_exec_statistic_u64(out, "Spills", "Number of spill instructions", (stats)->spills); \
   vk_add_exec_statistic_u64(out, "Fills", "Number of fill instructions", (stats)->fills); \
} while(0)

struct bifrost_stats {
   int32_t instrs;
   int32_t tuples;
   int32_t clauses;
   float cycles;
   float arith;
   float t;
   float ldst;
   float v;
   int16_t preloads;
   int16_t threads;
   int32_t code_size;
   int32_t loops;
   int32_t spills;
   int32_t fills;
};

static inline int
bifrost_stats_fprintf(FILE *fp, const char *prefix, const struct bifrost_stats *stats)
{
   return fprintf(fp, "%s shader: %u instrs, %u tuples, %u clauses, %f cycles, %f arith, %f t, %f ldst, %f v, %u preloads, %u threads, %u code size, %u loops, %u:%u spills:fills\n", prefix, stats->instrs, stats->tuples, stats->clauses, stats->cycles, stats->arith, stats->t, stats->ldst, stats->v, stats->preloads, stats->threads, stats->code_size, stats->loops, stats->spills, stats->fills);
}

static inline void
bifrost_stats_util_debug(struct util_debug_callback *debug, const char *prefix, const struct bifrost_stats *stats)
{
   util_debug_message(debug, SHADER_INFO, "%s shader: %u instrs, %u tuples, %u clauses, %f cycles, %f arith, %f t, %f ldst, %f v, %u preloads, %u threads, %u code size, %u loops, %u:%u spills:fills", prefix, stats->instrs, stats->tuples, stats->clauses, stats->cycles, stats->arith, stats->t, stats->ldst, stats->v, stats->preloads, stats->threads, stats->code_size, stats->loops, stats->spills, stats->fills);
}

#define vk_add_bifrost_stats(out, stats) do { \
   vk_add_exec_statistic_u64(out, "Instructions", "Instruction count", (stats)->instrs); \
   vk_add_exec_statistic_u64(out, "Tuples", "Tuple count", (stats)->tuples); \
   vk_add_exec_statistic_u64(out, "Clauses", "Clause count", (stats)->clauses); \
   vk_add_exec_statistic_f64(out, "Cycles", "Estimated normalized cycles", (stats)->cycles); \
   vk_add_exec_statistic_f64(out, "Arithmetic", "Estimated normalized arithmetic cycles", (stats)->arith); \
   vk_add_exec_statistic_f64(out, "Texture", "Estimated normalized Texture cycles", (stats)->t); \
   vk_add_exec_statistic_f64(out, "Load/store", "Estimated normalized Load/Store cycles", (stats)->ldst); \
   vk_add_exec_statistic_f64(out, "Varying", "Estimated normalized Varying cycles", (stats)->v); \
   vk_add_exec_statistic_u64(out, "Preloads", "Preload count", (stats)->preloads); \
   vk_add_exec_statistic_u64(out, "Threads", "Maximum number of threads in flight on a compute unit", (stats)->threads); \
   vk_add_exec_statistic_u64(out, "Code size", "Binary size in bytes", (stats)->code_size); \
   vk_add_exec_statistic_u64(out, "Loops", "Number of hardware loops", (stats)->loops); \
   vk_add_exec_statistic_u64(out, "Spills", "Number of spill instructions", (stats)->spills); \
   vk_add_exec_statistic_u64(out, "Fills", "Number of fill instructions", (stats)->fills); \
} while(0)

struct valhall_stats {
   int32_t instrs;
   float cycles;
   float fma;
   float cvt;
   float sfu;
   float v;
   float t;
   float ls;
   int32_t code_size;
   int16_t threads;
   int32_t loops;
   int32_t spills;
   int32_t fills;
};

static inline int
valhall_stats_fprintf(FILE *fp, const char *prefix, const struct valhall_stats *stats)
{
   return fprintf(fp, "%s shader: %u instrs, %f cycles, %f fma, %f cvt, %f sfu, %f v, %f t, %f ls, %u code size, %u threads, %u loops, %u:%u spills:fills\n", prefix, stats->instrs, stats->cycles, stats->fma, stats->cvt, stats->sfu, stats->v, stats->t, stats->ls, stats->code_size, stats->threads, stats->loops, stats->spills, stats->fills);
}

static inline void
valhall_stats_util_debug(struct util_debug_callback *debug, const char *prefix, const struct valhall_stats *stats)
{
   util_debug_message(debug, SHADER_INFO, "%s shader: %u instrs, %f cycles, %f fma, %f cvt, %f sfu, %f v, %f t, %f ls, %u code size, %u threads, %u loops, %u:%u spills:fills", prefix, stats->instrs, stats->cycles, stats->fma, stats->cvt, stats->sfu, stats->v, stats->t, stats->ls, stats->code_size, stats->threads, stats->loops, stats->spills, stats->fills);
}

#define vk_add_valhall_stats(out, stats) do { \
   vk_add_exec_statistic_u64(out, "Instructions", "Instruction count", (stats)->instrs); \
   vk_add_exec_statistic_f64(out, "Cycles", "Estimated normalized cycles", (stats)->cycles); \
   vk_add_exec_statistic_f64(out, "FMA", "Estimated normalized FMA (Fused Multiply-Add) cycles", (stats)->fma); \
   vk_add_exec_statistic_f64(out, "CVT", "Estimated normalized CVT (ConVerT) cycles", (stats)->cvt); \
   vk_add_exec_statistic_f64(out, "SFU", "Estimated normalized SFU (Special Function Unit) cycles", (stats)->sfu); \
   vk_add_exec_statistic_f64(out, "Varying", "Estimated normalized Varying cycles", (stats)->v); \
   vk_add_exec_statistic_f64(out, "Texture", "Estimated normalized Texture cycles", (stats)->t); \
   vk_add_exec_statistic_f64(out, "Load/store", "Estimated normalized Load/Store cycles", (stats)->ls); \
   vk_add_exec_statistic_u64(out, "Code size", "Binary size in bytes", (stats)->code_size); \
   vk_add_exec_statistic_u64(out, "Threads", "Maximum number of threads in flight on a compute unit", (stats)->threads); \
   vk_add_exec_statistic_u64(out, "Loops", "Number of hardware loops", (stats)->loops); \
   vk_add_exec_statistic_u64(out, "Spills", "Number of spill instructions", (stats)->spills); \
   vk_add_exec_statistic_u64(out, "Fills", "Number of fill instructions", (stats)->fills); \
} while(0)


enum panfrost_stat_isa {
   PANFROST_STAT_MIDGARD,
   PANFROST_STAT_BIFROST,
   PANFROST_STAT_VALHALL,
};

struct panfrost_stats {
   enum panfrost_stat_isa isa;
   union {
      struct midgard_stats midgard;
      struct bifrost_stats bifrost;
      struct valhall_stats valhall;
   };
};

#define vk_add_panfrost_stats(out, stats) do { \
    if ((stats)->isa == PANFROST_STAT_MIDGARD) \
       vk_add_midgard_stats(out, &(stats)->midgard); \
    if ((stats)->isa == PANFROST_STAT_BIFROST) \
       vk_add_bifrost_stats(out, &(stats)->bifrost); \
    if ((stats)->isa == PANFROST_STAT_VALHALL) \
       vk_add_valhall_stats(out, &(stats)->valhall); \
} while(0)

static inline void
panfrost_stats_fprintf(FILE *fp, const char *prefix, const struct panfrost_stats *stats)
{
    if (stats->isa == PANFROST_STAT_MIDGARD)
       midgard_stats_fprintf(fp, prefix, &stats->midgard);
    if (stats->isa == PANFROST_STAT_BIFROST)
       bifrost_stats_fprintf(fp, prefix, &stats->bifrost);
    if (stats->isa == PANFROST_STAT_VALHALL)
       valhall_stats_fprintf(fp, prefix, &stats->valhall);
}

static inline void
panfrost_stats_util_debug(struct util_debug_callback *debug, const char *prefix, const struct panfrost_stats *stats)
{
    if (stats->isa == PANFROST_STAT_MIDGARD)
       midgard_stats_util_debug(debug, prefix, &stats->midgard);
    if (stats->isa == PANFROST_STAT_BIFROST)
       bifrost_stats_util_debug(debug, prefix, &stats->bifrost);
    if (stats->isa == PANFROST_STAT_VALHALL)
       valhall_stats_util_debug(debug, prefix, &stats->valhall);
}


#endif

